/* 
 * Copyright (c) Comtrol Corporation <support@comtrol.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted prodived that the follwoing conditions
 * are met.
 * 1. Redistributions of source code must retain the above copyright 
 *    notive, this list of conditions and the following disclainer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials prodided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *       This product includes software developed by Comtrol Corporation.
 * 4. The name of Comtrol Corporation may not be used to endorse or 
 *    promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY COMTROL CORPORATION ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL COMTROL CORPORATION BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, LIFE OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD: /repoman/r/ncvs/src/sys/dev/rp/rpvar.h,v 1.5.2.1 2002/06/18 03:11:46 obrien Exp $
 */

/*
 * rpvar.h --- RocketPort data structure includes for FreeBSD
 */

#define RP_UNIT(x) dv_unit(x)
#define RP_PORT(x) (minor(x) & 0x3f)
#define MAX_RP_PORTS	128
#define	TIOCHST	_IOR('t', 125, int)	/* get all modem bits */

/**************************************************************************
  MUDBAC remapped for PCI
**************************************************************************/

#define _CFG_INT_PCI	0x40
#define _PCI_INT_FUNC	0x3A
#define _UPCI_INT_FUNC	0x4C

#define PCI_STROB	0x2000
#define UPCI_STROB_C	0x0000
#define UPCI_STROB_S	0x0001

#define INTR_EN_PCI	0x0010
#define INTR_EN_UPCI	0x0081

#define _UPCI_GPIO	0x54
#define _UPCI_QUAD_IN	0x0000
#define _UPCI_OCTAL_IN	0x4000

#define _UPCI_RING_IND 	0xc0


/* ISA-specific part of CONTROLLER_t */
struct ISACONTROLLER_T
{
    int MBaseIO;			/* rid of the Mudbac controller for this controller */
    int MReg0IO;			/* offset0 of the Mudbac controller for this controller */
    int MReg1IO;			/* offset1 of the Mudbac controller for this controller */
    int MReg2IO;			/* offset2 of the Mudbac controller for this controller */
    int MReg3IO;			/* offset3 of the Mudbac controller for this controller */
    unsigned char MReg2;
    unsigned char MReg3;
};
typedef struct ISACONTROLLER_T ISACONTROLLER_t;

#define ISACTL(ctlp) ((ISACONTROLLER_t *)((ctlp)->bus_ctlp))

#define ROCKET_C

// PCI IDs
#define RP_VENDOR_ID			0x11FE
#define RP_DEVICE_ID_32I		0x0001
#define RP_DEVICE_ID_8I			0x0002
#define RP_DEVICE_ID_16I		0x0003
#define RP_DEVICE_ID_4Q			0x0004
#define RP_DEVICE_ID_8O			0x0005
#define RP_DEVICE_ID_8J			0x0006
#define RP_DEVICE_ID_4J			0x0007
#define RP_DEVICE_ID_PL4		0x000A
#define RP_DEVICE_ID_PL8		0x000B
#define RP_DEVICE_ID_PL2		0x000E

// U designates UPCI
#define RP_DEVICE_ID_U32		0x0801
#define RP_DEVICE_ID_U8PL		0x0802
#define RP_DEVICE_ID_U16		0x0803
#define RP_DEVICE_ID_U8QO		0x0805
#define RP_DEVICE_ID_UP8QO		0x080B
#define RP_DEVICE_ID_UP2		0x080E
#define RP_DEVICE_ID_UP2SMPTE	0x080F
#define RP_DEVICE_ID_UP4J		0x0810
#define RP_DEVICE_ID_UP8J		0x0811
#define RP_DEVICE_ID_UP422QO	0x0812

// rocketmodems: U designates UPCI
// rm2 are here for identification but 
// will not operate without microcode 
// being loaded at runtime which this 
// driver does not support.
#define RP_DEVICE_ID_RM2_6		0x000C
#define RP_DEVICE_ID_RM2_4		0x000D
#define RP_DEVICE_ID_RM3_U8		0x080C
#define RP_DEVICE_ID_RM3_U4		0x080D
#define RP_DEVICE_ID_RM4_U8		0x0813
#define RP_DEVICE_ID_RM4_U4		0x0814

struct rp_port {
	struct tty *		rp_tty; /* cross reference */

	unsigned char		state;	/* state of dtr */

	int			rp_port;
	int			rp_flags;
	int			rp_unit:2;
	int			rp_aiop:2;
	int			rp_chan:3;
	int			rp_intmask;
	int			rp_imask; /* Input mask */
	int			rp_fifo_lw;
	int			rp_restart;
	int			rp_overflows;
	int			rp_rts_iflow:1;
	int			rp_disable_writes:1;
	int			rp_cts:1;
	int			rp_waiting:1;
	int			rp_xmit_stopped:1;
	CONTROLLER_t *		rp_ctlp;
	CHANNEL_t		rp_channel;
	unsigned short		TxBuf[TXFIFO_SIZE/2 + 1];
	unsigned short		RxBuf[RXFIFO_SIZE/2 + 1];
};

#define DTR_WAIT rp_tty->t_dtr_wait

Byte_t rp_sBitMapClrTbl[8] = {
                                 0xfe, 0xfd, 0xfb, 0xf7, 0xef, 0xdf, 0xbf, 0x7f
                             };

Byte_t rp_sBitMapSetTbl[8] = {
                                 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80
                             };

/* Actually not used */
#if notdef
extern struct termios deftermios;
#endif /* notdef */
